/* vim: set sts=4 ts=4 sw=4 et: */
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "stdconv.h"

typedef char *PCHAR;

class HtmlTree;
class HtmlTreeChild;
class myException;
class HtmlTree;
class WordCase;
void operateOnTree(HtmlTree *tree);
void putHtmlTreeInFile(char *targetFn, char *frameworkFn, char *defaultPageFn, HtmlTree *tree);
void HtmlTagSafePut(const char *sTextInfo, ofstream &ofsTargetFile);
void HtmlStringSafePut(const char *sTextInfo, ofstream &ofsTargetFile);
void treatMemoryAllocationFailure(void);

/**
 * General description of this program:
 * There are two main classes in this program: HtmlTree and LinkNode
 *
 * HtmlTree is a class used to handle trees.
 * HtmlTreeChild is a class used to handle children of the trees (not the root).
 *
 * Class HtmlTree handles such a structure:
 * A root node.
 * A series of children, who are also trees.
 *
 * Actual non-function members of class HtmlTree:
 * A pointer to the root node (HtmlTreeChild *).
 * A pointer pointing to the currently browsed tree node, also known as "active 
 * node". (HtmlTreeChild *)
 *
 * Class HtmlTree has the following operations:
 * Construction: Creates a new tree. The tree has only one node: the root.
 * Desstruction: Deletes the whole tree, including the root and the children 
 * trees.
 *
 * Sample tree:
 * C/C++ programming on DOS/Windows
 * +--MS-DOS Programming
 * |  +--Turbo C Programs
 * |  +--Quick C Programs
 * |  +--Turbo C++ Programs
 * |  \--Borland C++ Programs
 * \--Windows Programming
 *    +--16-bit Windows Programming
 *    |  +--Using Borland C++
 *    |  \--Using Microsoft C++
 *    \--32-bit Windows Programming
 *       +--Using Borland C++ Builder 5.0
 *       \--Using Microsoft Visual C++ 6.0
 */


/**
 * Class myException implements such a function: create an object that contains
 * a specific error type string and a specific error information string.
 *
 * This class is used for generating exception information.
 * A typical use of the class is described as follows.
 * In the throw statement: throw new myException("errorType", "an error occurred");
 * In the catch block:
 * catch(myException *E){
 *     cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
 *     delete E;
 * }// end catch
 */
class myException
{
public:
    myException(char *newType, char *newInfo);
    char *getErrorType(void);
    char *getErrorInfo(void);
protected:
    char type[16];
    char info[32];
};

/**
 * Class HtmlTree implements a multi-functional tree that holds hyper link 
 * information. It lets the user browse, modify the tree. It provides the user
 * overall modification functionality to the tree. It also provides some useful
 * features, such as internal clipboard, clearing the whole tree, etc. It 
 * protects operations in many aspects to avoid illegal operation.
 */
class HtmlTree
{
public:
    HtmlTree(void);
    HtmlTreeChild *_getRootNode(void);
    bool isNextExisting(void);
    bool isPrevExisting(void);
    bool isChildExisting(void);
    bool isRoot(void);
private:
    bool _isDescendant(HtmlTreeChild *testeeNode);
public:
    void browseNext(void);
    void browsePrev(void);
    void browseFirstChild(void);
    void browseParent(void);
    void browseRoot(void);
    void setText(char *newText);
    void setHyperRef(char *newHyperRef);
    char *getText(void);
    char *getHyperRef(void);
    void addAsChild(char *newText, char *newHyperRef);
    void addAfterCur(char *newText, char *newHyperRef);
    void addBeforeCur(char *newText, char *newHyperRef);
    void deleteOnlyChild(void);
    void deleteNext(void);
    void deleteFirstBrother(void);
    void deleteCur(void);
    void clear(void);
    void createSampleTree(void);
    void show(void);
    void pushActiveNode(void);
    void popActiveNode(void);
    void clearActiveNodeStack(void);
    void setHrefPrefix(char *newPrefix);
    void setHrefSuffix(char *newSuffix);
    char *getHrefPrefix(void);
    char *getHrefSuffix(void);
    bool isPrevActiveNode(void);
    void setBaseUrl(char *newBaseUrl);
    void setBaseTarget(char *newBaseTarget);
    char *getBaseUrl(void);
    char *getBaseTarget(void);
    void saveInFile(char *targetFileName);
private:
    void __saveInFile_SafePut(const char *sTextInfo, ofstream &ofsTargetFile);
public:
    void saveInFileVcl(char *targetFileName);
    void loadFromFile(char *sourceFileName);
private:
    void __getTextHref_loadFromFile(FILE *sourceFile, char *text, char *href);
public:
    void loadFromFileVcl(char *sourceFileName);
private:
    int __getTextHref_loadFromFileVcl(FILE *sourceFile, char *text, char *href);
public:
    void copyToClipboard(void);
private:
    void __checkDependencies_paste(void);
public:
    void pasteAfterCur(void);
    void pasteBeforeCur(void);
    void pasteAsChild(void);
    void setWorkPlace1(void);
    void setWorkPlace2(void);
    void gotoWorkPlace1(void);
    void gotoWorkPlace2(void);
private:
    void _setWorkPlace(int placeNum);
    void _gotoWorkPlace(int placeNum);
public:
    void setCSS(char *newcss);
    char *getCSS(void);
    bool isUsingParagraph(void);
    void setUsingParagraph(bool bUsing);
    ~HtmlTree(void);
private:
    HtmlTreeChild *rootNode;
    HtmlTreeChild *activeNode;
    HtmlTreeChild *activeNodeStack[32];
    int activeNodeStackFloater;
    char prefix[128];
    char suffix[128];
    char baseUrl[128];
    char baseTarget[128];
    HtmlTreeChild *clipboardNode;
    HtmlTreeChild *workPlace[2], *workPoint[2]; // a place is like a dir.; a point is like a file
    char css[320];
    bool bUsingParagraph;
};

/**
 * Class HtmlTreeChild is a subclass of class HtmlTree. It provides functions on
 * nodes of the HtmlTree. Each node of an HtmlTree is an HtmlTreeChild. That is 
 * to say, the relationship between HtmlTreeChild and HtmlTree is like the 
 * relationship between TreeView and TreeNode in Visual J++.
 */
class HtmlTreeChild
{
    friend class HtmlTree;
public:
    HtmlTreeChild(void);
    void addAsChild(char *newText, char *newHyperRef);
    void addAfterCur(char *newText, char *newHyperRef);
    void addBeforeCur(char *newText, char *newHyperRef);
private:
    void _setParent(HtmlTreeChild *newParent);
    void _setFirstChild(HtmlTreeChild *newChild);
    void _setPrev(HtmlTreeChild *prev);
    void _setNext(HtmlTreeChild *next);
public:
    void setText(char *newText);
    void setHyperRef(char *newHyperRef);
    char *getText(void);
    char *getHyperRef(void);
private:
    void __appendNode_addAsChild(HtmlTreeChild *newNode);
    HtmlTreeChild *_getNext(void);
    HtmlTreeChild *_getPrev(void);
    HtmlTreeChild *_getFirstChild(void);
    HtmlTreeChild *_getParent(void);
public:
    void deleteOnlyChild(void);
    void deleteNext(void);
    void deleteFirstBrother(void);
    ~HtmlTreeChild(void);
private:
    char text[64];
    HtmlTreeChild *parent; /* parent can also indicate whether the node is the root */
    HtmlTreeChild *prevNode;
    HtmlTreeChild *nextNode;
    HtmlTreeChild *firstChild;
    char hyperRef[128];
};

// Class WordCase implements extracting a string word by word automatically.
class WordCase
{
public:
    WordCase(void);
    WordCase(char *str);
    void setString(char *str);
    void resetIndex(void);
    WordCase &operator >> (char *outputString);
    WordCase &operator >> (int &intBuf);
    WordCase &operator >> (double &doubleBuf);
    ~WordCase(void);
protected:
    char *strPtr;
    int index;
    int strSize;
};

/**
 * Code for class myException
 */

/**
 * This is the constructor of class myException.
 */
myException::myException(char *newType, char *newInfo)
{
    strncpy(type, newType, 15);
    type[15] = '\0';
    strncpy(info, newInfo, 31);
    info[31] = '\0';
}

/**
 * This function returns the error type part of an exception.
 */
char *myException::getErrorType(void)
{
    return type;
}

/**
 * This function returns the error information part of an exception.
 */
char *myException::getErrorInfo(void)
{
    return info;
}

/**
 * Code for class HtmlTree
 */

/**
 * This is the contructor of class HtmlTree.
 * It initializes all data members in the class.
 * NOTE: activeNodeStack is not necessary to initialize because there is no
 * node available in the stack at first.
 */
HtmlTree::HtmlTree(void)
{
    rootNode = new HtmlTreeChild;
    activeNode = rootNode;
    activeNodeStackFloater = 0;
    prefix[0] = '\0';
    suffix[0] = '\0';
    baseUrl[0] = '\0';
    strcpy(baseTarget, "_blank");
    clipboardNode = NULL;
    workPlace[0] = NULL;
    workPlace[1] = NULL;
    workPoint[0] = NULL;
    workPoint[0] = NULL;
    css[0] = '\0';
    bUsingParagraph = true;
}

/**
 * This function adds a node as a child of the active node. All operations
 * on nodes in this class aim to the active node by default.
 *
 * This funtions calls function addAsChild in class HtmlTreeChild for the final
 * operation. By default, all adding and deleting functions in class HtmlTree
 * calls corresponding functions in class HtmlTreeChild.
 */
void HtmlTree::addAsChild(char *newText, char *newHyperRef = "")
{
    char tempBuf[382];
    if(newHyperRef[0]){
        /* if newHyperRef is not empty */
        strcpy(tempBuf, prefix);
        strcat(tempBuf, newHyperRef);
        strcat(tempBuf, suffix);
    }else{
        tempBuf[0] = '\0';
    }// end if
    activeNode->addAsChild(newText, tempBuf);
}

// adds a node after the active node
void HtmlTree::addAfterCur(char *newText, char *newHyperRef = "")
{
    char tempBuf[382];
    if(newHyperRef[0]){
        /* if newHyperRef is not empty */
        strcpy(tempBuf, prefix);
        strcat(tempBuf, newHyperRef);
        strcat(tempBuf, suffix);
    }else{
        tempBuf[0] = '\0';
    }// end if
    activeNode->addAfterCur(newText, tempBuf);
    browseNext(); // focus the added node
}

// adds a node before the active node
void HtmlTree::addBeforeCur(char *newText, char *newHyperRef = "")
{
    char tempBuf[382];
    if(newHyperRef[0]){
        /* if newHyperRef is not empty */
        strcpy(tempBuf, prefix);
        strcat(tempBuf, newHyperRef);
        strcat(tempBuf, suffix);
    }else{
        tempBuf[0] = '\0';
    }// end if
    activeNode->addBeforeCur(newText, tempBuf);
    browsePrev(); // focus the added node
}

/**
 * The following functions returns simple information of the HtmlTree.
 */
HtmlTreeChild *HtmlTree::_getRootNode(void)
{
    return rootNode;
}

bool HtmlTree::isNextExisting(void)
{
    return (activeNode->_getNext()) ? true : false;
}

bool HtmlTree::isPrevExisting(void)
{
    return (activeNode->_getPrev()) ? true : false;
}

bool HtmlTree::isChildExisting(void)
{
    return (activeNode->_getFirstChild()) ? true : false;
}

bool HtmlTree::isRoot(void)
{
    return (activeNode->_getParent()) ? false : true;
}

/**
 * This function determines if the node as argument is a
 * descendant of the current active node.
 *
 * This function is a relatively private function because HtmlTreeChild nodes
 * should not be exposed to the outside. This functions use functions in class
 * HtmlTreeChild to traverse the tree.
 *
 * parameters:
 * testeeNode - the node to be tested.
 *
 * return value:
 * true/false - whether the condition is true.
 */
bool HtmlTree::_isDescendant(HtmlTreeChild *testeeNode)
{
    HtmlTreeChild *searchPointer;
    searchPointer = activeNode;
    if(searchPointer == testeeNode){
        // the testee node is exactly pointed to by the active node
        return true;
    }// end if
    if(searchPointer->_getFirstChild() == NULL){
        // there is no child of searchPointer
        return false;
    }else{
        // there is at least one child
        searchPointer = searchPointer->_getFirstChild();
        while(true){
            if(searchPointer == testeeNode){
                // matching
                return true;
            }// end if
            if(searchPointer->_getFirstChild() != NULL){
                searchPointer = searchPointer->_getFirstChild();
            }else{
                while(searchPointer->_getNext() == NULL){
                    searchPointer = searchPointer->_getParent();
                    if(searchPointer == activeNode){
                        break; // break while
                    }// end if
                }// end while
                if(searchPointer == activeNode){
                    break; // break while
                }// end if
                searchPointer = searchPointer->_getNext();
            }// end if
        }// end while
        return false;
    }// end if
}

/**
 * The following functions are simple browsing functions that relocates the
 * active node.
 */
void HtmlTree::browseNext(void)
{
    if(activeNode->_getNext()){
        activeNode = activeNode->_getNext();
    }else{
        throw new myException("treeBrowse", "no next node");
    }// end if
}

void HtmlTree::browsePrev(void)
{
    if(activeNode->_getPrev()){
        activeNode = activeNode->_getPrev();
    }else{
        throw new myException("treeBrowse", "no previous node");
    }// end if
}

void HtmlTree::browseFirstChild(void)
{
    if(activeNode->_getFirstChild()){
        activeNode = activeNode->_getFirstChild();
    }else{
        throw new myException("treeBrowse", "no child node");
    }// end if
}

void HtmlTree::browseParent(void)
{
    if(activeNode->_getParent()){
        activeNode = activeNode->_getParent();
    }else{
        throw new myException("treeBrowse", "no parent node");
    }// end if
}

void HtmlTree::browseRoot(void)
{
    activeNode = rootNode;
}

/**
 * The following functions sets and gets information of the active node.
 */
void HtmlTree::setText(char *newText)
{
    activeNode->setText(newText);
}

void HtmlTree::setHyperRef(char *newHyperRef)
{
    activeNode->setHyperRef(newHyperRef);
}

char *HtmlTree::getText(void)
{
    return activeNode->getText();
}

char *HtmlTree::getHyperRef(void)
{
    return activeNode->getHyperRef();
}

/**
 * The following functions delete a node in three ways. They all provide a total
 * ability to delete any one node in the tree (except the root).
 */
void HtmlTree::deleteOnlyChild(void)
{
    activeNode->deleteOnlyChild();
}

void HtmlTree::deleteNext(void)
{
    activeNode->deleteNext();
}

void HtmlTree::deleteFirstBrother(void)
{
    activeNode->deleteFirstBrother();
}

/**
 * This functions deletes the current active node by calling the three functions
 * of deleting nodes. Current acitve node will be automatically set to a valid
 * node after the call of this function.
 */
void HtmlTree::deleteCur(void)
{
    if(!isRoot()){
        if(isPrevExisting()){
            browsePrev();
            deleteNext();
            if (isNextExisting()) {
                browseNext();
            }
        }else if(isNextExisting()){
            browseNext();
            deleteFirstBrother();
        }else{
            // this is not the root node
            browseParent();
            deleteOnlyChild();
        }// end if
    }else{
        throw new myException("deleteTreeNode", "cannot delete the root node");
    }// end if
}

/**
 * This function clears the whole tree. Child nodes are automatically deleted
 * by the destructor of class HtmlTreeChild.
 *
 * This function also clears other members of the tree.
 * For which members are to be cleared, refer to the declaration of the class.
 */
void HtmlTree::clear(void)
{
    browseRoot();
    if(isChildExisting()){
        browseFirstChild();
        while(isNextExisting()){
            deleteNext();
        }// end while
        browseParent();
        deleteOnlyChild();
    }// end if
    clearActiveNodeStack();
    setText("");
    setHyperRef("");
    setHrefPrefix("");
    setHrefSuffix("");
    setBaseUrl("");
    setBaseTarget("");
    clipboardNode = NULL;
    setCSS("");
    setUsingParagraph(true);
}

/**
 * This function creates a sample tree.
 */
void HtmlTree::createSampleTree(void)
{
    /* This function can be run only when the tree is empty */
    clear(); // clear the tree first
    try{
        setText("My Favorite Websites");
        addAsChild("International websites");
        browseFirstChild();
        addAsChild("Microsoft Corporate Website", "http://www.microsoft.com/");
        addAsChild("Borland/Inprise Website", "http://www.borland.com/");
        addAsChild("Ziff-Davis Network", "http://www.zdnet.com/");
        addAsChild("CNET Technology", "http://www.cnet.com/");
        addAsChild("Symantec Corporation Website", "http://www.symantec.com/");
        addAsChild("YAHOO!", "http://www.yahoo.com/");
        addAsChild("Robbie Mosaic's Homepage", "http://www.angelfire.com/wizard/robbiemosaic/");
        addAfterCur("Chinese websites");
        addAsChild("Eastday", "http://www.eastday.com/");
        addAsChild("Sina China", "http://www.sina.com.cn/");
        addAsChild("Sohu", "http://www.sohu.com/");
        addAsChild("NetEase", "http://www.163.com/");
        addAsChild("Chinaren", "http://www.chinaren.com/");
        addAfterCur("Sites within Shanghai Online VIP users' reach");
        addAsChild("PCHome", "http://www.pchome.net/");
        addAsChild("Kali Info", "http://www.kali.com.cn/");
        addAsChild("Shanghai Online", "http://www.online.sh.cn/");
        browseParent();
        setBaseTarget("frame2");
    }// end try
    catch(myException *E){
        cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
        delete E;
    }// end catch
}

/**
 * This function displays the tree.
 */
void HtmlTree::show(void)
{
    int layerBuffer[32];
    int layerFloater;
    int i;
    int linesDisplayed = 0;
    bool isLastNode;

    try{
        pushActiveNode();
    }// end try
    catch(myException *E){
        delete E;
        throw new myException("showTree", "active node stack full");
    }// end catch
    browseRoot();
    /* output the root node first */
    cout << activeNode->getText();
    if(isPrevActiveNode()){
        cout << '*' << endl;
    }else{
        cout << endl;
    }// end if
    if((++linesDisplayed) % 24 == 0){
        // a screen has been displayed; query the user
        if(!waitForMore()){
            // stop displaying
            popActiveNode();
            return;
        }// end if
    }// end if
    if(isChildExisting()){
        /* output the rest node if exist */
        browseFirstChild();
        layerFloater = 0;
        while(true){
            /* show a node */
            for(i = 0; i < layerFloater; i++){
                if(layerBuffer[i] == 0){
                    cout << "   ";
                }else{
                    cout << "|  ";
                }// end if
            }// end for
            if(isNextExisting()){
                isLastNode = false;
            }else{
                isLastNode = true;
            }// end if
            cout << (isLastNode ? "\\--" : "+--") << activeNode->getText();
            if(isPrevActiveNode()){
                cout << '*' << endl;
            }else{
                cout << endl;
            }// end if
            if((++linesDisplayed) % 24 == 0){
                // a screen has been displayed; query the user
                if(!waitForMore()){
                    // user break; stop displaying
                    popActiveNode();
                    return;
                }// end if
            }// end if
            if(isChildExisting()){
                /* browse the first child */
                browseFirstChild();
                layerBuffer[layerFloater] = isLastNode ? 0 : 1;
                layerFloater++;
            }else{
                /* no child can be browsed, test if it's the last node */
                while(!isNextExisting()){
                    /**
                     * Try to go to the next node; If no next node, 
                     * try again with the parent
                     */
                    browseParent();
                    if(isRoot()){
                        break; // break while
                    }// end if
                    layerFloater--;
                }// end while
                if(isRoot()){
                    break; // break while
                }
                browseNext();
            }// end if
        }// end while
    }// end if
    popActiveNode();
}

/**
 * Pushes the current active node into the active node stack.
 */
void HtmlTree::pushActiveNode(void)
{
    if(activeNodeStackFloater == 32){
        throw new myException("pushActiveNode", "stack is full");
    }else{
        activeNodeStack[activeNodeStackFloater] = activeNode;
        activeNodeStackFloater++;
    }// end if
}

/**
 * Pops one active node from the active node stack.
 *
 * This function checks for the validity of the active node at the top of
 * the stack first to avoid illegal access of memory.
 */
void HtmlTree::popActiveNode(void)
{
    if(activeNodeStackFloater == 0){
        throw new myException("popActiveNode", "stack is empty");
    }else{
        activeNodeStackFloater--;
        // use _isDescendant to check the validity of the active node to be popped
        activeNode = rootNode;
        if(_isDescendant(activeNodeStack[activeNodeStackFloater])){
            // valid
            activeNode = activeNodeStack[activeNodeStackFloater];
        }else{
            // do nothing to leave active node pointing to the root
        }// end if
    }// end if
}

// This function clears the active node stack.
void HtmlTree::clearActiveNodeStack(void)
{
    activeNodeStackFloater = 0;
}

/**
 * The following functions sets and gets text information of the active node.
 */
void HtmlTree::setHrefPrefix(char *newPrefix)
{
    strncpy(prefix, newPrefix, 128);
    prefix[127] = '\0';
}

void HtmlTree::setHrefSuffix(char *newSuffix)
{
    strncpy(suffix, newSuffix, 128);
    suffix[127] = '\0';
}

char *HtmlTree::getHrefPrefix(void)
{
    return prefix;
}

char *HtmlTree::getHrefSuffix(void)
{
    return suffix;
}

/**
 * This function checks whether the active node at the top of the active node 
 * stack equals to the current one.
 */
bool HtmlTree::isPrevActiveNode(void)
{
    if(activeNodeStackFloater > 0){
        if(activeNodeStack[activeNodeStackFloater - 1] == activeNode){
            return true;
        }else{
            return false;
        }// end if
    }else{
        throw new myException("checkActiveNode", "active node stack empty");
    }// end if
}

/**
 * The following functions sets and gets base URL and base target information
 * of the HtmlTree.
 */
void HtmlTree::setBaseUrl(char *newBaseUrl)
{
    strncpy(baseUrl, newBaseUrl, 128);
    baseUrl[127] = '\0';
}

void HtmlTree::setBaseTarget(char *newBaseTarget)
{
    strncpy(baseTarget, newBaseTarget, 128);
    baseTarget[127] = '\0';
}

char *HtmlTree::getBaseUrl(void)
{
    return baseUrl;
}

char *HtmlTree::getBaseTarget(void)
{
    return baseTarget;
}

/**
 * This function saves the tree in a file specified by argument targetFileName.
 *
 * parameters:
 * targetFileName - the name of the file to save the tree information to.
 */
void HtmlTree::saveInFile(char *targetFileName)
{
    ofstream targetFile(targetFileName);
    if(targetFile.fail()){
        throw new myException("treeSave", "error opening file");
    }// end if
    try{
        pushActiveNode();
    }// end try
    catch(myException *E){
        delete E;
        throw new myException("treeSave", "active node stack full");
    }// end try
    // start from the root node
    browseRoot();
    __saveInFile_SafePut(getText(), targetFile);
    __saveInFile_SafePut(getHyperRef(), targetFile);
    if(isChildExisting()){
        // deal with the children
        browseFirstChild();
        targetFile << 'c' << endl;
        while(true){
            __saveInFile_SafePut(getText(), targetFile);
            __saveInFile_SafePut(getHyperRef(), targetFile);
            if(isChildExisting()){
                // go one layer deeper
                browseFirstChild();
                targetFile << 'c' << endl;
            }else{
                // try to access the next node
                while(!isNextExisting()){
                    browseParent();
                    targetFile << 'p' << endl;
                    if(isRoot()){
                        break; // break while
                    }// end if
                }// end while
                if(isRoot()){
                    break; // break while
                }// end if
                browseNext();
                targetFile << 'n' << endl;
            }// end if
        }// end while
    }// end if
}

void HtmlTree::__saveInFile_SafePut(const char *sTextInfo, ofstream &ofsTargetFile) {
    int i;
    for (i = 0; sTextInfo[i] != '\0'; i++) {
        if (sTextInfo[i] >= ' ') {
            ofsTargetFile << sTextInfo[i];
        }
    }
    ofsTargetFile << endl;
}

/**
 * This function saves the tree in a file specified by argument targetFileName.
 * The output file format is compatible with Borland Visual Component Library.
 *
 * parameters:
 * targetFileName - the name of the file to save the tree information to.
 */
void HtmlTree::saveInFileVcl(char *targetFileName)
{
    ofstream targetFile(targetFileName);
    if(targetFile.fail()){
        throw new myException("treeSave", "error opening file");
    }// end if
    try{
        pushActiveNode();
    }// end try
    catch(myException *E){
        delete E;
        throw new myException("treeSave", "active node stack full");
    }// end try
    // start from the root node
    int indLvl = 0; /* indent level */
    int i;
    browseRoot();
    __saveInFile_SafePut(getText(), targetFile);
    __saveInFile_SafePut(getHyperRef(), targetFile);
    if(isChildExisting()){
        // deal with the children
        browseFirstChild();
        indLvl++;
        while(true){
            for (i = 0; i < indLvl; i++) targetFile << "\t";
            __saveInFile_SafePut(getText(), targetFile);
            for (i = 0; i < indLvl; i++) targetFile << "\t";
            __saveInFile_SafePut(getHyperRef(), targetFile);
            if(isChildExisting()){
                // go one layer deeper
                browseFirstChild();
                indLvl++;
            }else{
                // try to access the next node
                while(!isNextExisting()){
                    browseParent();
                    indLvl--;
                    if(isRoot()){
                        break; // break while
                    }// end if
                }// end while
                if(isRoot()){
                    break; // break while
                }// end if
                browseNext();
            }// end if
        }// end while
    }// end if
}

/**
 * This function loads the tree from a file specified by argument targetFileName.
 *
 * parameters:
 * targetFileName - the name of the file to load the tree information from.
 */
void HtmlTree::loadFromFile(char *sourceFileName)
{
    char strBuf[128], strBuf2[128];
    FILE *sourceFile;

    if((sourceFile = fopen(sourceFileName, "r")) == NULL){
        throw new myException("treeLoad", "error opening file");
    }// end if

    // delete the existing tree first
    clear();

    // start from the root node
    __getTextHref_loadFromFile(sourceFile, strBuf, strBuf2);
    setText(strBuf);
    setHyperRef(strBuf2);
    if(fileGetLine(strBuf, 128, sourceFile) != NULL){
        // there is at least one child node
        // verify the mark
        if(strBuf[0] != 'c'){
            throw new myException("treeLoad", "unknown file info");
        }// end if
        __getTextHref_loadFromFile(sourceFile, strBuf, strBuf2);
        addAsChild(strBuf, strBuf2);
        browseFirstChild();
        // deal with the other children
        while(true){
            if(fileGetLine(strBuf, 128, sourceFile) == NULL){
                throw new myException("treeLoad", "unexpected end-of-file");
            }// end if
            if(strBuf[0] == 'c'){
                // add a new child node
                __getTextHref_loadFromFile(sourceFile, strBuf, strBuf2);
                addAsChild(strBuf, strBuf2);
                browseFirstChild();
            }else if(strBuf[0] == 'n'){
                // add a new next node
                __getTextHref_loadFromFile(sourceFile, strBuf, strBuf2);
                addAfterCur(strBuf, strBuf2);
            }else if(strBuf[0] == 'p'){
                // browse parent
                browseParent();
                if(isRoot()){
                    break; // break while
                }// end if
            }else{
                // bad indicator
                throw new myException("treeLoad", "unexpected file info");
            }// end if
        }// end while
    }// end if
}

// This is a sub-function of function HtmlTree::loadFromFile().
void HtmlTree::__getTextHref_loadFromFile(FILE *sourceFile, char *text, char *href)
{
    if(fileGetLine(text, 128, sourceFile) == NULL){
        throw new myException("treeLoad", "unexpected end-of-file");
    }else if(fileGetLine(href, 128, sourceFile) == NULL){
        throw new myException("treeLoad", "unexpected end-of-file");
    }// end if
}

/**
 * This function loads the tree from a file specified by argument targetFileName.
 * Laws from Borland Visual Component Library applies to the file format.
 *
 * parameters:
 * targetFileName - the name of the file to load the tree information from.
 */
void HtmlTree::loadFromFileVcl(char *sourceFileName)
{
    char strBuf[128], strBuf2[128];
    FILE *sourceFile;

    if((sourceFile = fopen(sourceFileName, "r")) == NULL){
        throw new myException("treeLoad", "error opening file");
    }// end if

    // delete the existing tree first
    clear();

    // start from the root node
    int indLvl, pIndLvl; // indent level and preserved indent level
    int i;
    __getTextHref_loadFromFileVcl(sourceFile, strBuf, strBuf2);
    setText(strBuf);
    setHyperRef(strBuf2);
    indLvl = __getTextHref_loadFromFileVcl(sourceFile, strBuf, strBuf2);
    if(indLvl != -1){
        // there is at least one child node
        // verify the format
        if(indLvl < 1){
            // unexpected file info
            cout << "errA" << endl;
            throw new myException("treeLoad", "unknown file info");
        }// end if
        addAsChild(strBuf, strBuf2);
        browseFirstChild();
        // deal with the other children
        while(true){
            pIndLvl = indLvl;
            indLvl = __getTextHref_loadFromFileVcl(sourceFile, strBuf, strBuf2);
            if (indLvl == -1) {
                // EOF encountered
                browseRoot();
                break; // exit while
            } else if(pIndLvl < indLvl){
                // add a new child node
                addAsChild(strBuf, strBuf2);
                browseFirstChild();
            }else if(pIndLvl >= indLvl){
                // add a new next node
                for (i = 0; i < pIndLvl - indLvl; i++) browseParent();
                addAfterCur(strBuf, strBuf2);
            }
        }// end while
    }// end if
}

// This is a sub-function of function HtmlTree::loadFromFileVcl().
int HtmlTree::__getTextHref_loadFromFileVcl(FILE *sourceFile, char *text, char *href)
{
    char textBuf[128], hrefBuf[128];
    if(fileGetLine(textBuf, 128, sourceFile) == NULL){
        return -1;
    }else if(fileGetLine(hrefBuf, 128, sourceFile) == NULL){
        return -1;
    }// end if
    int indLvl;
    int i;
    for (i = 0; textBuf[i] == '\t'; i++);
    indLvl = i;
    for (i = 0; hrefBuf[i] == '\t'; i++);
    if (indLvl != i) {
        // the indent level of the text is not equal to the indent level of the href
        cout << "errB" << endl;
        throw new myException("treeLoad", "unknown file info");
    }
    for (i = 0; textBuf[indLvl + i] != '\0'; i++) {
        text[i] = textBuf[indLvl + i];
    }
    text[i] = '\0';
    for (i = 0; hrefBuf[indLvl + i] != '\0'; i++) {
        href[i] = hrefBuf[indLvl + i];
    }
    href[i] = '\0';
    return indLvl;
}

// Records the active node in the internal clipboard.
void HtmlTree::copyToClipboard(void)
{
    clipboardNode = activeNode;
}

// Checks dependencies for paste operation. This is a sub-function for pasting.
void HtmlTree::__checkDependencies_paste(void)
{
    HtmlTreeChild *backupPointer;
    if(clipboardNode == NULL){
        throw new myException("treePaste", "clipboard is empty");
    }// end if
    /**
     * Push the active node into the stack, and check the validity
     * of the node in clipboard.
     */
    try{
        pushActiveNode();
    }// end try
    catch(myException *E){
        delete E;
        throw new myException("treePaste", "stack is full");
    }// end catch
    browseRoot();
    if(!_isDescendant(clipboardNode)){
        clipboardNode = NULL;
        popActiveNode();
        throw new myException("treePaste", "the node to paste is invalid");
    }// end if
    // now check if the current node is a descendant of the node to paste
    popActiveNode();
    backupPointer = activeNode;
    activeNode = clipboardNode;
    if(_isDescendant(backupPointer) && activeNode != backupPointer){
        // do not change clipboard node because it is valid
        activeNode = backupPointer; // restore
        throw new myException("treePaste", "cannot paste parent to child");
    }// end if
    activeNode = backupPointer; // restore
}

/**
 * The following three functions pastes a subtree starting from the node pointed
 * to in the clipboard.
 */
void HtmlTree::pasteAfterCur(void)
{
    HtmlTreeChild *backupPointer, *idxPtr;
    // check dependencies
    __checkDependencies_paste();
    // start copying
    pushActiveNode(); // no need to check again
    backupPointer = clipboardNode;
    idxPtr = clipboardNode;
    addAfterCur("", ""); // add an empty node for accurate href assigment
    setText(idxPtr->getText());
    setHyperRef(idxPtr->getHyperRef());
    if(idxPtr->_getFirstChild()){
        addAsChild("", "");
        browseFirstChild();
        idxPtr = idxPtr->_getFirstChild();
        while(true){
            setText(idxPtr->getText());
            setHyperRef(idxPtr->getHyperRef());
            if(idxPtr->_getFirstChild()){
                addAsChild("", "");
                browseFirstChild();
                idxPtr = idxPtr->_getFirstChild();
            }else{
                while(!(idxPtr->_getNext())){
                    browseParent();
                    idxPtr = idxPtr->_getParent();
                    if(idxPtr == backupPointer){
                        // matches the subtree root
                        break; // break while
                    }// end if
                }// end while
                if(idxPtr == backupPointer){
                    break; // break while
                }// end if
                addAfterCur("", "");
                idxPtr = idxPtr->_getNext();
            }// end if
        }// end while
    }// end if
    popActiveNode();
    browseNext();
}

void HtmlTree::pasteBeforeCur(void)
{
    HtmlTreeChild *backupPointer, *idxPtr;
    // check dependencies
    __checkDependencies_paste();
    // start copying
    pushActiveNode(); // no need to check again
    backupPointer = clipboardNode;
    idxPtr = clipboardNode;
    addBeforeCur("", ""); // add an empty node for accurate href assigment
    setText(idxPtr->getText());
    setHyperRef(idxPtr->getHyperRef());
    if(idxPtr->_getFirstChild()){
        addAsChild("", "");
        browseFirstChild();
        idxPtr = idxPtr->_getFirstChild();
        while(true){
            setText(idxPtr->getText());
            setHyperRef(idxPtr->getHyperRef());
            if(idxPtr->_getFirstChild()){
                addAsChild("", "");
                browseFirstChild();
                idxPtr = idxPtr->_getFirstChild();
            }else{
                while(!(idxPtr->_getNext())){
                    browseParent();
                    idxPtr = idxPtr->_getParent();
                    if(idxPtr == backupPointer){
                        // matches the subtree root
                        break; // break while
                    }// end if
                }// end while
                if(idxPtr == backupPointer){
                    break; // break while
                }// end if
                addAfterCur("", "");
                idxPtr = idxPtr->_getNext();
            }// end if
        }// end while
    }// end if
    popActiveNode();
    browsePrev();
}

void HtmlTree::pasteAsChild(void)
{
    HtmlTreeChild *backupPointer, *idxPtr;
    // check dependencies
    __checkDependencies_paste();
    // start copying
    pushActiveNode(); // no need to check again
    backupPointer = clipboardNode;
    idxPtr = clipboardNode;
    addAsChild("", ""); // add an empty node for accurate href assigment
    browseFirstChild();
    // go to the node
    while( activeNode->_getNext() != NULL )
    {
        browseNext();
    }
    setText(idxPtr->getText());
    setHyperRef(idxPtr->getHyperRef());
    if( idxPtr->_getFirstChild() )
    {
        addAsChild("", "");
        browseFirstChild();
        idxPtr = idxPtr->_getFirstChild();
        while( true )
        {
            setText(idxPtr->getText());
            setHyperRef(idxPtr->getHyperRef());
            if( idxPtr->_getFirstChild() )
            {
                addAsChild("", "");
                browseFirstChild();
                idxPtr = idxPtr->_getFirstChild();
            }
            else
            {
                while( idxPtr->_getNext() == NULL )
                {
                    browseParent();
                    idxPtr = idxPtr->_getParent();
                    if( idxPtr == backupPointer )
                    { // matches the subtree root
                        break; // break while
                    } // end if
                } // end while
                if( idxPtr == backupPointer )
                {
                    break; // break while
                }// end if
                addAfterCur("", "");
                idxPtr = idxPtr->_getNext();
            } // end if
        } // end while
    } // end if
    popActiveNode();
}

/**
 * The following four functions are unfinished.
 */

void HtmlTree::setWorkPlace1(void)
{
    _setWorkPlace(1);
}

void HtmlTree::setWorkPlace2(void)
{
    _setWorkPlace(2);
}

void HtmlTree::gotoWorkPlace1(void)
{
    _gotoWorkPlace(1);
}

void HtmlTree::gotoWorkPlace2(void)
{
    _gotoWorkPlace(2);
}

void HtmlTree::_setWorkPlace(int placeNum)
{
    int i = placeNum - 1;
    if( !isRoot() )
    {
        workPlace[i] = activeNode->_getParent();
    }
    else
    {
        workPlace[i] = rootNode;
    }
    workPoint[i] = activeNode;
}

void HtmlTree::_gotoWorkPlace(int placeNum)
{
    int i;
    HtmlTreeChild *preserver = activeNode;
    if( !isRoot() )
    {
        // update working point if possible
        for( i = 0; i < 2; i++ )
        {
            if( workPlace[i] == activeNode->_getParent() )
            {
                workPoint[i] = activeNode;
                break;
            }
        }
    }
    // goto the working place
    i = placeNum - 1;
    browseRoot();
    if( _isDescendant(workPlace[i]) )
    {
        activeNode = workPlace[i];
    }
    else
    { // working place not found
        activeNode = preserver;
        workPlace[i] = NULL;
        throw new myException("treeGotoWorkPlc", "Working place not found");
    }
    // find the working point
    if( isChildExisting() )
    {
        browseFirstChild();
        while( activeNode != workPoint[i] && isNextExisting() )
        {
            browseNext();
        }
        if( activeNode != workPoint[i] )
        { // working point not found
            activeNode = preserver;
            workPoint[i] = NULL;
            throw new myException("treeGotoWorkPlc", "Working point not found");
        }
    }
    else
    {
        // working point not found
        activeNode = preserver;
        workPoint[i] = NULL;
        throw new myException("treeGotoWorkPlc", "Working point not found");
    }
}

void HtmlTree::setCSS(char *newcss)
{
    int i;
    for( i = 0; newcss[i] && i < 319; i++ )
    {
        css[i] = newcss[i];
    }
    css[i] = '\0';
}

char *HtmlTree::getCSS(void)
{
    return css;
}

bool HtmlTree::isUsingParagraph(void)
{
    return bUsingParagraph;
}

void HtmlTree::setUsingParagraph(bool bUsing)
{
    bUsingParagraph = bUsing;
}

/**
 * This is the destructor of class HtmlTree. It frees the dynamically 
 * allocated node rootNode.
 */
HtmlTree::~HtmlTree(void)
{
    /* clean allocated objects */
    delete rootNode;
}

/**
 * Code for class HtmlTreeChild
 */

/**
 * This is the contructor of class HtmlTreeChild.
 * It initializes all data members in the class.
 */
HtmlTreeChild::HtmlTreeChild(void)
{
    text[0] = '\0';
    parent = NULL;
    prevNode = NULL;
    nextNode = NULL;
    firstChild = NULL;
    hyperRef[0] = '\0';
}

/**
 * Adds a new node as a child.
 * The function modifies the current node when there is no child node.
 * The function calls HtmlTreeChild::_appendNode() to finish the operation,
 * if there is at least one child node.
 */
void HtmlTreeChild::addAsChild(char *newText, char *newHyperRef = "")
{
    HtmlTreeChild *temp;
    temp = new HtmlTreeChild;
    // only parentNode is needed to be set, other values were automatically set empty
    temp->_setParent(this);
    temp->setText(newText);
    temp->setHyperRef(newHyperRef);
    /* establish the structure */
    if(firstChild){
        /* use appendNode function */
        firstChild->__appendNode_addAsChild(temp);
    }else{
        this->_setFirstChild(temp);
    }/* end if */
}

// This is a sub-function of function HtmlTreeChild::addAsChild().
void HtmlTreeChild::__appendNode_addAsChild(HtmlTreeChild *newNode)
{
    /* NOTE: Pointer "this" should be valid */
    if(nextNode == NULL){
        /* add the node after this */
        newNode->_setPrev(this);
        this->_setNext(newNode);
    }else{
        nextNode->__appendNode_addAsChild(newNode);
    }/* end if */
}

/**
 * This function adds a node after the current node.
 * This action affects: the current node and the node after current if it exists.
 */
void HtmlTreeChild::addAfterCur(char *newText, char *newHyperRef = "")
{
    
    if(this->parent){
        /* current node is not the root */
        HtmlTreeChild *temp;
        temp = new HtmlTreeChild;
        temp->_setParent(parent);
        temp->_setPrev(this);
        temp->_setNext(nextNode);
        temp->_setFirstChild(NULL);
        temp->setText(newText);
        temp->setHyperRef(newHyperRef);
        if(nextNode){
            // if the next node exists, set the previous node to temp
            nextNode->_setPrev(temp);
        }// end if
        this->_setNext(temp);
    }else{
        throw new myException("addTreeNode", "root node has no brother");
    }// end if
}

/**
 * This function adds a node after the current node.
 * This action affects: the current node and the parent/previous node.
 */
void HtmlTreeChild::addBeforeCur(char *newText, char *newHyperRef = "")
{
    if(this->parent){
        /* current node is not the root */
        HtmlTreeChild *temp;
        temp = new HtmlTreeChild;
        /* common assignments */
        temp->_setParent(parent);
        temp->_setPrev(prevNode);
        temp->_setNext(this);
        temp->_setFirstChild(NULL);
        temp->setText(newText);
        temp->setHyperRef(newHyperRef);
        if(prevNode == NULL){
            /* this is the first node, parent is needed to be modified */
            /* temp->prevNode needn't be modified */
            parent->_setFirstChild(temp);
        }else{
            prevNode->_setNext(temp);
        }/* end if */
        this->_setPrev(temp);
    }else{
        throw new myException("addTreeNode", "root node has no brother");
    }// end if
}

/**
 * This function assigns a parent to the current node.
 *
 * This function is an internal function that should not be used outside
 * the class. While creating this function, I thought there is need to modify
 * properties of another node during the operation of the current. I think
 * although C++ is a object-oriented language that allows functions to access
 * properties of objects of the same class, it is still not good to observe
 * the rule of smallTalk (limit access to current object only) because it is
 * like our human. If a function have full access to another person, that is
 * unreasonable.
 *
 * The following functions sets properties of the current node.
 */
void HtmlTreeChild::_setParent(HtmlTreeChild *newParent)
{
    parent = newParent;
}

void HtmlTreeChild::_setFirstChild(HtmlTreeChild *newChild)
{
    firstChild = newChild;
}

void HtmlTreeChild::_setPrev(HtmlTreeChild *prev)
{
    prevNode = prev;
}

void HtmlTreeChild::_setNext(HtmlTreeChild *next)
{
    nextNode = next;
}

/**
 * The following functions sets and gets the text and the href of the current 
 * node.
 */
void HtmlTreeChild::setText(char *newText)
{
    strncpy(text, newText, 63);
    text[63] = '\0';
}

void HtmlTreeChild::setHyperRef(char *newHyperRef)
{
    strncpy(hyperRef, newHyperRef, 127);
    hyperRef[127] = '\0';
}

char *HtmlTreeChild::getText(void)
{
    return text;
}

char *HtmlTreeChild::getHyperRef(void)
{
    return hyperRef;
}

/**
 * The following functions are internal functions that gets properties of
 * the current node.
 */
HtmlTreeChild *HtmlTreeChild::_getNext(void)
{
    return nextNode;
}

HtmlTreeChild *HtmlTreeChild::_getPrev(void)
{
    return prevNode;
}

HtmlTreeChild *HtmlTreeChild::_getFirstChild(void)
{
    return firstChild;
}

HtmlTreeChild *HtmlTreeChild::_getParent(void)
{
    return parent;
}

/**
 * Deletes the only child of the current node
 * This action affects: the current node
 */
void HtmlTreeChild::deleteOnlyChild(void)
{
    if(firstChild){
        if(firstChild->nextNode == NULL){
            /* delete the child */
            delete firstChild;
            firstChild = NULL;
        }else{
            throw new myException("deleteTreeNode", "no only child");
        }// end if
    }else{
        throw new myException("deleteTreeNode", "no child");
    }// end if
}

/**
 * Deletes the only child of the current node
 * This action affects: the current node
 */
void HtmlTreeChild::deleteNext(void)
{
    // this action affects: the current node and the node after next if it exists
    if(nextNode){
        /* next node exists */
        HtmlTreeChild *temp;
        temp = nextNode->_getNext();
        nextNode->_setNext(NULL); /* avoid deleting the rest in destruction */
        delete nextNode;
        this->_setNext(temp);
        if(temp){
            temp->_setPrev(this);
        }// end if
    }else{
        throw new myException("deleteTreeNode", "no next node");
    }// end if
}

/**
 * Deletes the only child of the current node
 * This action affects: the parent node and the current node if it is the
 * next node of the deleted node
 */
void HtmlTreeChild::deleteFirstBrother(void)
{
    if(prevNode){
        /* this is not the first child */
        HtmlTreeChild *temp;
        temp = parent->firstChild;
        parent->_setFirstChild(temp->nextNode);
        temp->_setNext(NULL); /* avoid deleting the rest in destruction */
        delete temp;
        parent->firstChild->_setPrev(NULL);
    }else{
        throw new myException("deleteTreeNode", "first brother is current");
    }// end if
}

/**
 * This is the destructor of class HtmlTreeChild. It frees its next node 
 * and first child automatically. If the rest link structure is perfect, 
 * it will free all the rest link. If the child tree structure is perfect, 
 * it will free all the subtrees.
 */
HtmlTreeChild::~HtmlTreeChild(void)
{
    if(nextNode){
        delete nextNode;
    }// end if
    if(firstChild){
        delete firstChild;
    }// end if
}

/**
 * Code for class WordCAse
 */

/**
 * The default constructor of class WordCase. It initializes the members.
 */
WordCase::WordCase(void)
{
    strPtr = NULL;
    index = 0;
    strSize = 0;
}

/**
 * The constructor of class WordCase. It initializes the members according
 * to the specific argument "str".
 *
 * parameter:
 * str - the string to point to as the source
 */
WordCase::WordCase(char *str)
{
    strPtr = str;
    index = 0;
    strSize = strlen(strPtr) + 1;
}

/**
 * Sets the members according to the specific argument "str".
 *
 * parameter:
 * str - the string to point to as the source
 */
void WordCase::setString(char *str)
{
    strPtr = str;
    index = 0;
    strSize = strlen(strPtr) + 1;
}

/**
 * Resets the index. The index is a member that indicates where the next
 * extraction will begin.
 */
void WordCase::resetIndex(void)
{
    index = 0;
}

/**
 * Each of the following functions extracts one word. The word can 
 * be a number or a string. If the extraction fails, an exception 
 * will be thrown.
 */
WordCase &WordCase::operator >> (char *outputString)
{
    char ch;
    int i;

    // check dependencies
    if(strPtr == NULL){
        // uninitialized wordcase
        throw new myException("wCaseOutputStr", "wordcase not initialized");
    }// end if

    // skip the blank first
    for(; ch = strPtr[index], (ch == ' ' || ch == '\t') && ch != '\0'; index++){
        ; // do nothing
    }// end for
    for(i = 0; ; i++, index++){
        ch = strPtr[index];
        if(ch == ' ' || ch == '\t' || ch == '\0'){
            break; // break for
        }// end if
        outputString[i] = ch;
    }// end for
    outputString[i] = '\0';
    if(i == 0){
        // nothing has been extracted, because end-of-string encountered
        throw new myException("wCaseOutputStr", "end-of-string encountered");
    }// end if
    return *this;
}

WordCase &WordCase::operator >> (int &intBuf)
{
    char *strBuf;
    char *endPtr;
    int preIndex;
    
    // check dependencies
    if(strPtr == NULL){
        // uninitialized wordcase
        throw new myException("wCaseOutputNum", "wordcase not initialized");
    }// end if
    // start the conversion
    strBuf = new char[strSize];
    // save the current index before extracting
    preIndex = index;
    try{
        operator >> (strBuf);
    }// end try
    catch(myException *E){
        delete E;
        delete [] strBuf;
        throw new myException("wCaseOutputNum", "error extracting from string");
    }// end catch
    intBuf = strtol(strBuf, &endPtr, 10);
    if(endPtr == strBuf){
        // nothing has been extracted
        delete [] strBuf;
        // restore the index if failed to extract
        index = preIndex;
        throw new myException("wCaseOutputNum", "no number extractable");
    }// end if
    delete [] strBuf;
    return *this;
}

WordCase &WordCase::operator >> (double &doubleBuf)
{
    char *strBuf;
    char *endPtr;
    int preIndex;
    
    // check dependencies
    if(strPtr == NULL){
        // uninitialized wordcase
        throw new myException("wCaseOutputNum", "wordcase not initialized");
    }// end if
    // start the conversion
    strBuf = new char[strSize];
    // save the current index
    preIndex = index;
    try{
        operator >> (strBuf);
    }// end try
    catch(myException *E){
        delete E;
        delete [] strBuf;
        throw new myException("wCaseOutputNum", "error extracting from string");
    }// end catch
    doubleBuf = strtod(strBuf, &endPtr);
    if(endPtr == strBuf){
        // nothing has been extracted
        delete [] strBuf;
        // restore the index if failed to extract
        index = preIndex;
        throw new myException("wCaseOutputNum", "no number extractable");
    }// end if
    delete [] strBuf;
    return *this;
}

// This is the destructor of class WordCase. Nothing to manually free.
WordCase::~WordCase(void)
{
    // do nothing
}

/**
 * Code within no class starts here
 */

void main(int argc, char *argv[])
{
    // set exception handler for operator "new"
    set_new_handler(treatMemoryAllocationFailure);

    HtmlTree *defaultTree = new HtmlTree();
    char targetFilename[128], frmwrkFilename[128], defPageFilename[128];

    if(argc < 2){
        cout <<
        "Usage:\n" <<
        "HTMLTREE.EXE output_html_filename [output_framework_filename]\n" <<
        "[default_page_filename]\n";
        return;
    }else{
        cout << "The filename for output is \"" << argv[1] << "\".\n";
    }/* end if */
    strcpy(targetFilename, argv[1]);
    if(argc >= 3){
        defaultTree->setBaseTarget("frame2");
        strcpy(frmwrkFilename, argv[2]);
    }else{
        strcpy(frmwrkFilename, "");
    }// end if
    if(argc >= 4){
        strcpy(defPageFilename, argv[3]);
    }else{
        strcpy(defPageFilename, "");
    }// end if
    operateOnTree(defaultTree);
    try{
        putHtmlTreeInFile(targetFilename, frmwrkFilename, defPageFilename, defaultTree);
    }// end try
    catch(myException *E){
        cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
    }// end catch
}

void operateOnTree(HtmlTree *tree)
{
    char strBuf[128], strBuf2[128];
    char cmdBuf[128];
    char css[320];
    int i;
    int ch;
    WordCase wordCase;

    /**
     * Command list:
     *
     * aac - add as a child
     * aann - add as the next node
     * aapn - add as the previous node
     * del - delete the current node
     * bfc - browse the first child
     * bnn - browse the next node
     * bpn - browse the previous node
     * bparn - browse the parent node
     * list - list the tree
     * info - show current node information (text and hyper reference)
     * settext - set current node text
     * sethref - set current node HyperRef
     * setprefix - set hyper reference prefix
     * setsuffix - set hyper reference suffix
     * setbu - set base URL
     * setbt - set base target
     * save/vcl - save the tree in a file
     * load/vcl - load a tree from a file
     * zap - delete the whole tree
     * push - push the active node into the stack
     * pop - pop one active node from the stack
     * copy - copy the active node to the clipboard
     * pasbef - paste the node in the clipboard before the current
     * pasaft - paste the node in the clipboard after the current
     * sample - create a sample tree
     * setcss - set CSS attributes for the web page
     * setusingpara - choose whether to use <p> or <div>
     * help - display help information
     * quit - quit editing
     */
    while(true){
        cout << '-';
        getLine(strBuf, 128);
        try{
            wordCase.setString(strBuf);
            wordCase >> cmdBuf;
        }// end try
        catch(myException *E){
            delete E;
            cmdBuf[0] = '\0'; // empty command line
        }// end catch
        if(strcmp(cmdBuf, "aac") == 0){
            // add as a child
            cout << "Text: ";
            getLine(strBuf, 128);
            cout << "HRef: ";
            getLine(strBuf2, 128);
            try{
                tree->addAsChild(strBuf, strBuf2);
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "aann") == 0){
            // add as the next node
            cout << "Text: ";
            getLine(strBuf, 128);
            cout << "HRef: ";
            getLine(strBuf2, 128);
            try{
                tree->addAfterCur(strBuf, strBuf2);
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "aapn") == 0){
            // add as the previous node
            cout << "Text: ";
            getLine(strBuf, 128);
            cout << "HRef: ";
            getLine(strBuf2, 128);
            try{
                tree->addBeforeCur(strBuf, strBuf2);
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "del") == 0){
            // delete currently browsed node
            try{
                tree->deleteCur();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "bfc") == 0){
            // browse the first child
            try{
                tree->browseFirstChild();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "bnn") == 0){
            // browse the next node
            try{
                tree->browseNext();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "bpn") == 0){
            // browse the previous node
            try{
                tree->browsePrev();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "bparn") == 0){
            // browse the parent node
            try{
                tree->browseParent();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "list") == 0){
            // list the tree
            try{
                tree->show();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "getinfo") == 0){
            // show the text and the href of the currently browsed node
            cout << tree->getText() << endl;
            cout << "href=\'" << tree->getHyperRef() << "\'\n";
        }else if(strcmp(cmdBuf, "settext") == 0){
            // set current node text
            cout << "current text: " << tree->getText() << endl;
            cout << "new text: ";
            getLine(strBuf, 128);
            tree->setText(strBuf);
        }else if(strcmp(cmdBuf, "sethref") == 0){
            // set current node href
            cout << "current href: " << tree->getHyperRef() << endl;
            cout << "new href: ";
            getLine(strBuf, 128);
            tree->setHyperRef(strBuf);
        }else if(strcmp(cmdBuf, "setprefix") == 0){
            // set prefix
            cout << "current prefix: " << tree->getHrefPrefix() << endl;
            cout << "new prefix: ";
            getLine(strBuf, 128);
            tree->setHrefPrefix(strBuf);
        }else if(strcmp(cmdBuf, "setsuffix") == 0){
            // set suffix
            cout << "current suffix: " << tree->getHrefSuffix() << endl;
            cout << "new suffix: ";
            getLine(strBuf, 128);
            tree->setHrefSuffix(strBuf);
        }else if(strcmp(cmdBuf, "setbu") == 0){
            // set base URL
            cout << "current base URL: " << tree->getBaseUrl() << endl;
            cout << "new base URL: ";
            getLine(strBuf, 128);
            tree->setBaseUrl(strBuf);
        }else if(strcmp(cmdBuf, "setbt") == 0){
            // set base target
            cout << "current base target: " << tree->getBaseTarget() << endl;
            cout << "new base target: ";
            getLine(strBuf, 128);
            tree->setBaseTarget(strBuf);
        }else if(strcmp(cmdBuf, "save") == 0 || strcmp(cmdBuf, "savevcl") == 0){
            // save in file
            cout << "File name: ";
            getLine(strBuf, 128);
            try{
                if (strcmp(cmdBuf, "save") == 0) {
                    tree->saveInFile(strBuf);
                } else {
                    tree->saveInFileVcl(strBuf);
                }
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "load") == 0 || strcmp(cmdBuf, "loadvcl") == 0){
            // load from file
            cout << "This action will delete the whole tree.\n";
            cout << "Are you sure? [Y/N] ";
            getLine(strBuf, 128);
            if(strBuf[0] == 'y' || strBuf[0] == 'Y'){
                cout << "File name: ";
                getLine(strBuf, 128);
                try{
                    if (strcmp(cmdBuf, "load") == 0) {
                        tree->loadFromFile(strBuf);
                    } else {
                        tree->loadFromFileVcl(strBuf);
                    }
                }// end try
                catch(myException *E){
                    cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                    delete E;
                }// end catch
            }// end if
        }else if(strcmp(cmdBuf, "zap") == 0){
            cout << "You are about to delete the whole tree.\n";
            cout << "Are you sure? [Y/N] ";
            getLine(strBuf, 128);
            if(strBuf[0] == 'y' || strBuf[0] == 'Y'){
                tree->clear();
            }// end if
        }else if(strcmp(cmdBuf, "push") == 0){
            try{
                tree->pushActiveNode();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "pop") == 0){
            try{
                tree->popActiveNode();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "copy") == 0){
            tree->copyToClipboard();
        }else if(strcmp(cmdBuf, "pasbef") == 0){
            try{
                tree->pasteBeforeCur();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "pasaft") == 0){
            try{
                tree->pasteAfterCur();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "paschd") == 0){
            try{
                tree->pasteAsChild();
            }// end try
            catch(myException *E){
                cout << "Error of " << E->getErrorType() << ": " << E->getErrorInfo() << endl;
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "setwp1") == 0){
            tree->setWorkPlace1();
        }else if(strcmp(cmdBuf, "setwp2") == 0){
            tree->setWorkPlace2();
        }else if(strcmp(cmdBuf, "gotowp1") == 0){
            try
            {
                tree->gotoWorkPlace1();
            }
            catch( myException *e )
            {
                cout << "Error of " << e->getErrorType() << ": " << e->getErrorInfo() << endl;
                delete e;
            }
        }else if(strcmp(cmdBuf, "gotowp2") == 0){
            try
            {
                tree->gotoWorkPlace2();
            }
            catch( myException *e )
            {
                cout << "Error of " << e->getErrorType() << ": " << e->getErrorInfo() << endl;
                delete e;
            }
        }else if(strcmp(cmdBuf, "move12") == 0){
            try
            {
                tree->gotoWorkPlace1();
                tree->copyToClipboard();
                tree->gotoWorkPlace2();
                tree->pasteAfterCur();
                tree->gotoWorkPlace1();
                tree->deleteCur();
            }
            catch( myException *e )
            {
                cout << "Error of " << e->getErrorType() << ": " << e->getErrorInfo() << endl;
                delete e;
            }
        }else if(strcmp(cmdBuf, "sample") == 0){
            cout << "This action will delete the whole tree.\n";
            cout << "Are you sure? [Y/N] ";
            getLine(strBuf, 128);
            if(strBuf[0] == 'y' || strBuf[0] == 'Y'){
                tree->createSampleTree();
                tree->show();
            }// end if
        }else if(strcmp(cmdBuf, "setcss") == 0){
            cout <<
            "Example CSS style table: " << endl <<
            ". { font-family: Tahoma, Verdana, MS Sans Serif; font-size: 8pt; color: silver }" << endl <<
            "body { background-color: darkblue }" << endl <<
            "a { color: silver }" << endl <<
            "a:hover { color: white; background-color: violet }" << endl << endl;
            cout << "Current CSS style table: " << endl;
            cout << tree->getCSS() << endl;
            cout << "Please enter new style table (^Z to end): " << endl;
            for( i = 0, ch = getc(stdin); i < 319 && ch != -1; i++, ch = getc(stdin) )
            {
                css[i] = ch;
            }
            css[i] = '\0';
            fflush(stdin);
            tree->setCSS(css);
        }else if(strcmp(cmdBuf, "setusingpara") == 0){
            cout << "Use paragraph <p> or division <div>? [P/D] ";
            switch( getchar() )
            {
            case 'p':
            case 'P':
                tree->setUsingParagraph(true);
                break;
            case 'd':
            case 'D':
                tree->setUsingParagraph(false);
                break;
            default:
                cout << "Error entering!" << endl;
            }
            fflush(stdin);
        }else if(strcmp(cmdBuf, "help") == 0 || strcmp(cmdBuf, "?") == 0){
            // show help
            try{
                wordCase >> cmdBuf;
                // if successful
                if(strcmp(cmdBuf, "aac") == 0){
                    cout <<
                    "Adds a new node as a child node of the currently browsed node." << endl <<
                    "You will be prompted to enter the text and href information " << endl <<
                    "for the node." << endl <<
                    "NOTE: The node will be added after all existing child nodes " << endl <<
                    "of the currently browsed node." << endl;
                }else if(strcmp(cmdBuf, "aann") == 0){
                    cout <<
                    "Adds a new node as the next node of the currently browsed node." << endl <<
                    "You will be prompted to enter the text and href information " << endl <<
                    "for the node." << endl <<
                    "NOTE: The root node cannot be added brother to." << endl;
                }else if(strcmp(cmdBuf, "aapn") == 0){
                    cout <<
                    "Adds a new node as the previous node of the currently browsed node." << endl <<
                    "You will be prompted to enter the text and href information " << endl <<
                    "for the node." << endl <<
                    "NOTE: The root node cannot be added brother to." << endl;
                }else if(strcmp(cmdBuf, "del") == 0){
                    cout <<
                    "Deletes the currently browsed node." << endl <<
                    "NOTE: The root node cannot be deleted." << endl;
                }else if(strcmp(cmdBuf, "bfc") == 0){
                    cout <<
                    "Locates the first child node of the currently browsed node." << endl <<
                    "NOTE: This requires that at least one child node exists." << endl;
                }else if(strcmp(cmdBuf, "bnn") == 0){
                    cout <<
                    "Locates the next node of the currently browsed node." << endl <<
                    "NOTE: This requires the existence of the next node." << endl;
                }else if(strcmp(cmdBuf, "bpn") == 0){
                    cout <<
                    "Locates the previous node of the currently browsed node." << endl <<
                    "NOTE: This requires the existence of the previous node." << endl;
                }else if(strcmp(cmdBuf, "bparn") == 0){
                    cout <<
                    "Locates the parent node of the currently browsed node." << endl <<
                    "NOTE: This requires the existence of the parent node." << endl;
                }else if(strcmp(cmdBuf, "list") == 0){
                    cout <<
                    "Lists the tree." << endl;
                }else if(strcmp(cmdBuf, "getinfo") == 0){
                    cout <<
                    "Shows the text and the href of the currently browsed node." << endl;
                }else if(strcmp(cmdBuf, "settext") == 0){
                    cout <<
                    "Shows the text of the currently browsed node and lets you modify it." << endl;
                }else if(strcmp(cmdBuf, "sethref") == 0){
                    cout <<
                    "Shows the href of the currently browsed node and lets you modify it." << endl;
                }else if(strcmp(cmdBuf, "setprefix") == 0){
                    cout <<
                    "Shows and lets you set the href prefix." << endl <<
                    "NOTE: The prefix will be automatically added every time you create a new node." << endl;
                }else if(strcmp(cmdBuf, "setsuffix") == 0){
                    cout <<
                    "Shows and lets you set the href suffix." << endl <<
                    "NOTE: The suffix will be automatically added every time you create a new node." << endl;
                }else if(strcmp(cmdBuf, "setbu") == 0){
                    cout <<
                    "Shows and lets you set the base URL." << endl <<
                    "NOTE: The base URL will affect all hyper references in the whole document." << endl <<
                    "It will be a virtual prefix to all hyper references in the document." << endl;
                }else if(strcmp(cmdBuf, "setbt") == 0){
                    cout <<
                    "Shows and lets you set the base target." << endl <<
                    "NOTE: The base target will affect all <A> tags in the document.\n" << 
                    "It will make hyper links opened in the target window by default.\n";
                }else if(strcmp(cmdBuf, "save") == 0 || strcmp(cmdBuf, "savevcl") == 0){
                    cout <<
                    "Saves the tree in a file.\n" << 
                    "You will be prompted for the name of the file.\n" << 
                    "\"savevcl\" supports VCL format.\n";
                }else if(strcmp(cmdBuf, "load") == 0 || strcmp(cmdBuf, "loadvcl") == 0){
                    cout <<
                    "Loads a tree from a file.\n" << 
                    "You will be prompted for the name of the file.\n" << 
                    "\"loadvcl\" supports VCL format.\n";
                }else if(strcmp(cmdBuf, "zap") == 0){
                    cout <<
                    "Deletes the whole tree.\n";
                }else if(strcmp(cmdBuf, "help") == 0){
                    cout <<
                    "Displays helpful information.\n" << 
                    "You can use it in the format \"help <command>\".\n";
                }else if(strcmp(cmdBuf, "push") == 0){
                    cout <<
                    "Pushes the current active node into the stack.\n";
                }else if(strcmp(cmdBuf, "pop") == 0){
                    cout <<
                    "Pops previously pushed active node from the stack.\n";
                }else if(strcmp(cmdBuf, "copy") == 0){
                    cout <<
                    "Copies the active node to the clipboard.\n";
                }else if(strcmp(cmdBuf, "pasbef") == 0){
                    cout <<
                    "Pastes the node in the clipboard before the current.\n";
                }else if(strcmp(cmdBuf, "pasaft") == 0){
                    cout <<
                    "Pastes the node in the clipboard after the current.\n";
                }else if(strcmp(cmdBuf, "paschd") == 0){
                    cout <<
                    "Pastes the node in the clipboard as a child of the current.\n";
                }else if(strcmp(cmdBuf, "setwp1") == 0){
                    cout <<
                    "Saves current working place in wp1.\n";
                }else if(strcmp(cmdBuf, "setwp2") == 0){
                    cout <<
                    "Saves current working place in wp2.\n";
                }else if(strcmp(cmdBuf, "gotowp1") == 0){
                    cout <<
                    "Goes to working place 1.\n";
                }else if(strcmp(cmdBuf, "gotowp2") == 0){
                    cout <<
                    "Goes to working place 2.\n";
                }else if(strcmp(cmdBuf, "move12") == 0){
                    cout <<
                    "Moves a node in from wp1 to wp2.\n" << 
                    "It is a batch of gotowp1, copy, gotowp2, pasaft, gotowp1, del\n";
                }else if(strcmp(cmdBuf, "sample") == 0){
                    cout <<
                    "Creates a sample tree. All information in the tree you are editing will be\n" << 
                    "cleared.\n";
                }else if(strcmp(cmdBuf, "setcss") == 0){
                    cout <<
                    "Sets the style table for the web page\n";
                }else if(strcmp(cmdBuf, "setusingpara") == 0){
                    cout <<
                    "Sets whether the program uses <p> or <div> as default tags.\n";
                }else{
                    cout <<
                    "There is not such a command that matchs your request.\n";
                }// end if
            }// end try
            catch(myException *E){
                cout <<
                "Command list:\n" << 
                "\n" << 
                "aac - add as child\n" << 
                "aann - add as the next node\n" << 
                "aapn - add as the previous node\n" 
                "del - delete current node\n" << 
                "bfc - browse first child\n" << 
                "bnn - browse the next node\n" << 
                "bpn - browse the previous node\n" << 
                "bparn - browse the parent node\n" << 
                "list - list the tree\n" << 
                "getinfo - show current node information (text and hyper reference)\n" << 
                "settext - set current node text\n" << 
                "sethref - set current node href\n" << 
                "setprefix - set hyper reference prefix\n" << 
                "setsuffix - set hyper reference suffix\n" << 
                "setbu - set base URL\n" << 
                "setbt - set base target\n" << 
                "save/savevcl - save the tree in a file\n" << 
                "load/loadvcl - load a tree from a file\n" << 
                "zap - delete the whole tree\n" << 
                "push - push the active node into the stack\n" << 
                "pop - pop one active node from the stack\n";
                if(!waitForMore()) goto helpInfoEnd;
                cout <<
                "copy - copy the active node to the clipboard\n" << 
                "pasbef - paste the node in the clipboard before the current\n" << 
                "pasaft - paste the node in the clipboard after the current\n" << 
                "paschd - paste the node in the clipboard as a child\n" << 
                "setwp1/setwp2 - set working place 1 / 2\n" << 
                "gotowp1/gotowp2 - go to working place 1 / 2\n" << 
                "move12 - moves a node from wp1 to wp2\n" << 
                "sample - create a sample tree\n" << 
                "setcss - sets the style table\n" << 
                "setusingpara - choose whether to use <p> or <div>\n" << 
                "help / ? - display help information\n" << 
                "quit - quit\n";
            helpInfoEnd:
                delete E;
            }// end catch
        }else if(strcmp(cmdBuf, "quit") == 0){
            break; // break while
        }else if(cmdBuf[0]){
            // a bad command was given
            cout << "Bad command\n";
        }else{
            // because cmdBuf is empty, do nothing
        }// end if
    }// end while
}

void putHtmlTreeInFile(char *targetFn, char *frameworkFn, char *defaultPageFn, HtmlTree *tree)
{
    int i;
    int layerBuffer[32];
    int layerFloater;
    int curNumber;
    char lineString[128];
    char compBuf[32]; // buffer for string comparison
    ifstream tempHead("templateHead.tem");
    ifstream tempEnd("templateEnd.tem");
    ifstream tempFrame("templateFrame.tem");
    ofstream targetFile(targetFn);
    char tagString[16];

    // try to open the files
    if(tempHead.fail() || tempEnd.fail() || tempFrame.fail() || targetFile.fail()){
        throw new myException("generateOutput", "error opening file(s)");
    }// end if
    tree->pushActiveNode();
    tree->browseRoot();
    // create the framework if required
    if(frameworkFn[0] != '\0'){
        // create the framework
        ofstream framework(frameworkFn);
        if(framework.fail()){
            throw new myException("generateOutput", "error opening file(s)");
        }// end if
        while(!tempFrame.eof()){
            tempFrame.getline(lineString, 128);
            strncpy(compBuf, lineString, 31);
            compBuf[31] = '\0';
            strlwr(compBuf); // transform the string into lower case
            if(strstr(compBuf, "<title>") == compBuf){
                /* The line is the title line, which will be changed */
                framework << "<title>" << tree->getText() << "</title>\n";
            }else if(strstr(compBuf, "<frame name=frame1") == compBuf){
                // the left page will be set
                framework << "<frame name=frame1 src=\'" << targetFn << "\'>\n";
            }else if(defaultPageFn[0] != '\0' &&
                     strncmp(compBuf,
                             "<frame name=frame2",
                             strlen("<frame name=frame2")) == 0){
                // the default page will be set
                framework <<
                "<frame name=frame2 src=\'" << defaultPageFn << "\'>\n";
            }else{
                framework << lineString << endl;
            }// end if
        }// end while
    }// end if

    /**
     * From here is the creation process of the target file.
     */
    // initialization
    strcpy(tagString, tree->isUsingParagraph() ? "p" : "div");
    // create the head first
    while(!tempHead.eof()){
        tempHead.getline(lineString, 128);
        strncpy(compBuf, lineString, 31);
        compBuf[31] = '\0';
        strlwr(compBuf);
        if(strstr(compBuf, "<title>") == compBuf){
            /* The line is the title line, which will be changed */
            targetFile << "<title>";
            HtmlTagSafePut(tree->getText(), targetFile);
            targetFile << "</title>\n";
        }else if( strstr(compBuf, ". {") == compBuf ){
            // put CSS style table
            HtmlStringSafePut(tree->getCSS(), targetFile);
        }else{
            targetFile << lineString << endl;
        }// end if
    }// end while
    targetFile << "<base";
    if( tree->getBaseUrl()[0] != '\0' )
    {
        targetFile << " href=\'";
        HtmlStringSafePut(tree->getBaseUrl(), targetFile);
        targetFile << "\'";
    }
    targetFile << " target=\'";
    HtmlStringSafePut(tree->getBaseTarget(), targetFile);
    targetFile << "\'>\n";

    // create the body
    layerFloater = 0;
    curNumber = 1;
    if(tree->isChildExisting()){
        tree->browseFirstChild();
        while(true){
            /**
             * Each tag is in this format:
             * <p> (may be <div>)
             * <span id=paragraph[n1]-[n2]-[n3][...]>[text]</span>
             * </p> (may be </div>)
             *
             * Notice that n1, n2, ... starts from 1.
             * If there is a hyper reference, it will be automatically added.
             */
            // targetFile << "<" << tagString << " style=\'margin-left: " << layerFloater * 2 << "em\'>\n";
            targetFile << "<" << tagString << ">\n";
            targetFile << "<span id=paragraph";
            for(i = 0; i < layerFloater; i++){
                targetFile << layerBuffer[i] << "-";
            }// end for
            targetFile << curNumber << ">";
            if(tree->getHyperRef()[0] == '\0'){
                HtmlStringSafePut(tree->getText(), targetFile);
            }else{
                targetFile << "<a href=\'";
                HtmlStringSafePut(tree->getHyperRef(), targetFile);
                targetFile << "\'>";
                HtmlTagSafePut(tree->getText(), targetFile);
                targetFile << "</a>";
            }// end if
            targetFile << "</span>\n";
            targetFile << "</" << tagString << ">\n";
            // begin transportation
            if(tree->isChildExisting()){
                /**
                 * Output another tag
                 * <div id=paragraph[n1]-[n2]-[n3][...]sub 
                 * style="margin-left: 2em">
                 */
                targetFile << "<div id=paragraph";
                for(i = 0; i < layerFloater; i++){
                    targetFile << layerBuffer[i] << "-";
                }// end for
                targetFile << curNumber << "sub" << " style=\"margin-left: 2em";
                targetFile << "\">\n";
                /* go to the first child node */
                tree->browseFirstChild();
                layerBuffer[layerFloater] = curNumber;
                curNumber = 1;
                layerFloater++;
                /**
                 * NOTE: There is an unresolved problem:
                 * The treatment of a too deep tree.
                 * In that occasion, layerFloater will exceed the allowed range.
                 */
            }else{
                /* try whether there is a next node */
                while( !tree->isNextExisting() ){
                    layerFloater--;
                    if(layerFloater == -1){
                        break; // break while
                    }// end if
                    /**
                     * Output tag
                     * </div>
                     */
                    targetFile << "</div>\n";
                    tree->browseParent();
                    curNumber = layerBuffer[layerFloater]; // follow the change
                }// end while
                if(layerFloater == -1){
                    break; // break while
                }// end if
                curNumber++;
                tree->browseNext();
            }// end if
        }// end while
    }// end if

    // create the last part
    tree->popActiveNode();
    while(!tempEnd.eof()){
        tempEnd.getline(lineString, 128);
        targetFile << lineString << endl;
    }// end while
}

void HtmlTagSafePut(const char *sTextInfo, ofstream &ofsTargetFile) {
    int i;
    for (i = 0; sTextInfo[i] != '\0'; i++) {
        if (sTextInfo[i] >= ' ') {
            if (sTextInfo[i] == '>') {
                ofsTargetFile << "&gt;";
            } else if (sTextInfo[i] == '<') {
                ofsTargetFile << "&lt;";
            } else if (sTextInfo[i] == '&') {
                ofsTargetFile << "&amp;";
            } else {
                ofsTargetFile << sTextInfo[i];
            }
        }
    }
}

void HtmlStringSafePut(const char *sTextInfo, ofstream &ofsTargetFile) {
    int i;
    for (i = 0; sTextInfo[i] != '\0'; i++) {
        if (sTextInfo[i] >= ' ') {
            if (sTextInfo[i] == '\'') {
                ofsTargetFile << "&#39;";
            } else if (sTextInfo[i] == '&') {
                ofsTargetFile << "&amp;";
            } else {
                ofsTargetFile << sTextInfo[i];
            }
        }
        else if( sTextInfo[i] == '\n' )
        {
            ofsTargetFile << endl;
        }
    }
}

void treatMemoryAllocationFailure(void)
{
    throw "Memory overflow!";
}
